@@ -1,12 +1,12 @@
 <?php
 /**
- * Copyright 2013-2014 Horde LLC (http://www.horde.org/)
+ * Copyright 2013-2015 Horde LLC (http://www.horde.org/)
  *
  * See the enclosed file COPYING for license information (LGPL). If you
  * did not receive this file, see http://www.horde.org/licenses/lgpl21.
  *
  * @category  Horde
- * @copyright 2013-2014 Horde LLC
+ * @copyright 2013-2015 Horde LLC
  * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
  * @package   Smtp
  */
@@ -17,11 +17,14 @@
  * Implements the following SMTP-related RFCs:
  * <pre>
  *   - RFC 1870/STD 10: Message Size Declaration
+ *   - RFC 1985: SMTP Service Extension for Remote Message Queue Starting
  *   - RFC 2034: Enhanced-Status-Codes
  *   - RFC 2195: CRAM-MD5 (SASL Authentication)
  *   - RFC 2595/4616: TLS & PLAIN (SASL Authentication)
  *   - RFC 2831: DIGEST-MD5 authentication mechanism (obsoleted by RFC 6331)
  *   - RFC 2920/STD 60: Pipelining
+ *   - RFC 3030: SMTP Service Extensions for Transmission of Large and Binary
+ *               MIME Messages
  *   - RFC 3207: Secure SMTP over TLS
  *   - RFC 3463: Enhanced Mail System Status Codes
  *   - RFC 4422: SASL Authentication (for DIGEST-MD5)
@@ -38,7 +41,6 @@
  * <pre>
  *   - RFC 1845: CHECKPOINT
  *   - RFC 2852: DELIVERYBY
- *   - RFC 3030: BINARYMIME/CHUNKING
  *   - RFC 3461: DSN
  *   - RFC 3865: NO-SOLICITING
  *   - RFC 3885: MTRK
@@ -52,12 +54,14 @@
  *
  * @author    Michael Slusarz <slusarz@horde.org>
  * @category  Horde
- * @copyright 2013-2014 Horde LLC
+ * @copyright 2013-2015 Horde LLC
  * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
  * @package   Smtp
  *
  * @property-read boolean $data_8bit  Does server support sending 8-bit MIME
  *                                    data?
+ * @property-read boolean $data_binary  Does server support sending binary
+ *                                      MIME data? (@since 1.7.0)
  * @property-read boolean $data_intl  Does server support sending
  *                                    internationalized (UTF-8) header data?
  *                                    (@since 1.6.0)
@@ -66,6 +70,8 @@
  */
 class Horde_Smtp implements Serializable
 {
+    const CHUNK_DEFAULT = 1048576;
+
     /**
      * Connection to the SMTP server.
      *
@@ -113,61 +119,50 @@ class Horde_Smtp implements Serializable
      * Constructor.
      *
      * @param array $params   Configuration parameters:
-     * <ul>
-     *  <li>
-     *   debug: (string) If set, will output debug information to the stream
-     *          provided. The value can be any PHP supported wrapper that
-     *          can be opened via fopen().
-     *          DEFAULT: No debug output
-     *  </li>
-     *  <li>
-     *   host: (string) The SMTP server.
-     *         DEFAULT: localhost
-     *  </li>
-     *  <li>
-     *   password: (mixed) The SMTP password or a Horde_Smtp_Password object
-     *             (since 1.1.0).
-     *             DEFAULT: NONE
-     *  </li>
-     *  <li>
-     *   port: (string) The SMTP port.
-     *         DEFAULT: 587 (See RFC 6409/STD 72)
-     *  </li>
-     *  <li>
-     *   secure: (string) Use SSL or TLS to connect.
-     *           DEFAULT: true (use 'tls' option, if available)
-     *   <ul>
-     *    <li>false (No encryption)</li>
-     *    <li>'ssl' (Auto-detect SSL version)</li>
-     *    <li>'sslv2' (Force SSL version 3)</li>
-     *    <li>'sslv3' (Force SSL version 2)</li>
-     *    <li>'tls' (TLS; started via protocol-level negotation over
-     *    unencrypted channel; RECOMMENDED way of initiating secure
-     *    connection)
-     *    <li>'tlsv1' (TLS direct version 1.x connection to server) [@since
-     *    1.3.0]</li>
-     *    <li>true (Use TLS, if available) [@since 1.2.0]</li>
-     *   </ul>
-     *  </li>
-     *  <li>
-     *   timeout: (integer) Connection timeout, in seconds.
-     *            DEFAULT: 30 seconds
-     *  </li>
-     *  <li>
-     *   username: (string) The SMTP username.
-     *             DEFAULT: NONE
-     *  </li>
-     *  <li>
-     *   xoauth2_token: (string) If set, will authenticate via the XOAUTH2
-     *                  mechanism (if available) with this token. Either a
-     *                  string or a Horde_Smtp_Password object (since 1.1.0).
-     *  </li>
-     * </ul>
+     *   - chunk_size: (integer) If CHUNKING is supported on the server, the
+     *                 chunk size (in octets) to send. 0 will disable chunking.
+     *                 @since 1.7.0
+     *   - context: (array) Any context parameters passed to
+     *              stream_create_context(). @since 1.9.0
+     *   - debug: (string) If set, will output debug information to the stream
+     *            provided. The value can be any PHP supported wrapper that
+     *            can be opened via fopen().
+     *            DEFAULT: No debug output
+     *   - host: (string) The SMTP server.
+     *           DEFAULT: localhost
+     *   - localhost: (string) The hostname of the localhost. (since 1.9.0)
+     *                DEFAULT: Auto-determined.
+     *   - password: (mixed) The SMTP password or a Horde_Smtp_Password object
+     *               (since 1.1.0).
+     *               DEFAULT: NONE
+     *   - port: (string) The SMTP port.
+     *           DEFAULT: 587 (See RFC 6409/STD 72)
+     *   - secure: (string) Use SSL or TLS to connect.
+     *             DEFAULT: true (use 'tls' option, if available)
+     *     - false (No encryption)
+     *     - 'ssl' (Auto-detect SSL version)
+     *     - 'sslv2' (Force SSL version 3)
+     *     - 'sslv3' (Force SSL version 2)
+     *     - 'tls' (TLS; started via protocol-level negotation over
+     *       unencrypted channel; RECOMMENDED way of initiating secure
+     *       connection)
+     *     - 'tlsv1' (TLS direct version 1.x connection to server) [@since
+     *       1.3.0]
+     *     - true (Use TLS, if available) [@since 1.2.0]
+     *   - timeout: (integer) Connection timeout, in seconds.
+     *              DEFAULT: 30 seconds
+     *   - username: (string) The SMTP username.
+     *               DEFAULT: NONE
+     *   - xoauth2_token: (string) If set, will authenticate via the XOAUTH2
+     *                    mechanism (if available) with this token. Either a
+     *                    string or a Horde_Smtp_Password object (since 1.1.0).
      */
     public function __construct(array $params = array())
     {
         // Default values.
         $params = array_merge(array(
+            'chunk_size' => self::CHUNK_DEFAULT,
+            'context' => array(),
             'host' => 'localhost',
             'port' => 587,
             'secure' => true,
@@ -248,6 +243,10 @@ class Horde_Smtp implements Serializable
             // RFC 6152
             return $this->queryExtension('8BITMIME');
 
+        case 'data_binary':
+            // RFC 3030
+            return $this->queryExtension('BINARYMIME');
+
         case 'data_intl':
             // RFC 6531
             return $this->queryExtension('SMTPUTF8');
@@ -340,7 +339,7 @@ class Horde_Smtp implements Serializable
             return false;
         }
 
-        $ext = strtoupper($ext);
+        $ext = Horde_String::upper($ext);
 
         return isset($this->_extensions[$ext])
             ? $this->_extensions[$ext]
@@ -375,6 +374,7 @@ class Horde_Smtp implements Serializable
                     $this->getParam('port'),
                     $this->getParam('timeout'),
                     $this->getParam('secure'),
+                    $this->getParam('context'),
                     array(
                         'debug' => $this->_debug
                     )
@@ -395,7 +395,9 @@ class Horde_Smtp implements Serializable
             ));
 
             // Get initial line (RFC 5321 [3.1]).
-            $this->_getResponse(220, 'logout');
+            $this->_getResponse(220, array(
+                'error' => 'logout'
+            ));
         }
 
         $this->_hello();
@@ -483,15 +485,6 @@ class Horde_Smtp implements Serializable
      *                     Horde_Mail_Rfc822_List object, a string, or an
      *                     array of addresses.
      * @param mixed $data  The data to send. Either a stream or a string.
-     * @param array $opts  Additional options:
-     * <pre>
-     *   - 8bit: (boolean) If true, $data is a MIME message with arbitrary
-     *           octet content (i.e. 8-bit encoding).
-     *           DEFAULT: false
-     *   - intl: (boolean) If true, $data contains internationalized header
-     *           content (UTF-8). (@since 1.6.0)
-     *           DEFAULT: false
-     * </pre>
      *
      * @return array  If no receipients were successful, a
      *                Horde_Smtp_Exception will be thrown. If at least one
@@ -502,6 +495,7 @@ class Horde_Smtp implements Serializable
      *             or a Horde_Smtp_Exception (if messages was not accepted).
      *
      * @throws Horde_Smtp_Exception
+     * @throws Horde_Smtp_Exception_Recipients
      */
     public function send($from, $to, $data, array $opts = array())
     {
@@ -511,79 +505,157 @@ class Horde_Smtp implements Serializable
             $from = new Horde_Mail_Rfc822_Address($from);
         }
 
-        /* RFC 6531 */
-        if (!empty($opts['intl'])) {
-            if (!$this->data_intl) {
-                throw new InvalidArgumentException(
-                    'Server does not support sending internationalized header data.'
-                );
-            }
-
-            /* RFC 6531[1.2] requires 8BITMIME to be available. */
-            $opts['8bit'] = true;
-        }
-
-        /* RFC 6152[3] */
-        if (!empty($opts['8bit']) && !$this->data_8bit) {
-            throw new InvalidArgumentException(
-                'Server does not support sending 8-bit data.'
-            );
-        }
-
-        $mailcmd = 'MAIL FROM:<' .
-            (empty($opts['intl']) ? $from->bare_address_idn : $from->bare_address) .
-            '>';
-
-        // RFC 1870[6]
-        if ($this->queryExtension('SIZE')) {
-            if (is_resource($data)) {
-                fseek($data, 0, SEEK_END);
-                $size = ftell($data);
-            } else {
-                $size = strlen($data);
-            }
-
-            $mailcmd .= ' SIZE=' . intval($size);
-        }
-
-        // RFC 6531[3.4]
-        if (!empty($opts['intl'])) {
-            $mailcmd .= ' SMTPUTF8';
-        }
-
-        // RFC 6152[3]
-        if (!empty($opts['8bit'])) {
-            $mailcmd .= ' BODY=8BITMIME';
-        } elseif ($this->_debug->active && $this->data_8bit) {
-            /* Only output extended 7bit command if debug is active (it is
-             * default and does not need to be explicitly declared). */
-            $mailcmd .= ' BODY=7BIT';
-        }
-
-        $cmds = array($mailcmd);
+        /* Are EAI addresses present? */
+        $eai = $from->eai;
 
         if (!($to instanceof Horde_Mail_Rfc822_List)) {
             $to = new Horde_Mail_Rfc822_List($to);
         }
 
-        $recipients = empty($opts['intl'])
-            ? $to->bare_addresses_idn
-            : $to->bare_addresses;
+        if (!$eai) {
+            foreach ($to->raw_addresses as $val) {
+                if ($eai = $val->eai) {
+                    break;
+                }
+            }
+        }
+
+        /* RFC 6531: EAI */
+        if ($eai && !$this->data_intl) {
+            throw new InvalidArgumentException(
+                'Server does not support sending internationalized header data.'
+            );
+        }
+
+        $stream = fopen('php://temp', 'r+');
+
+        stream_filter_register('horde_smtp_data', 'Horde_Smtp_Filter_Data');
+        $filter_data = stream_filter_append(
+            $stream,
+            'horde_smtp_data',
+            STREAM_FILTER_WRITE
+        );
+
+        stream_filter_register('horde_smtp_body', 'Horde_Smtp_Filter_Body');
+        $filter_body_params = new stdClass;
+        $filter_body = stream_filter_append(
+            $stream,
+            'horde_smtp_body',
+            STREAM_FILTER_WRITE,
+            $filter_body_params
+        );
+
+        if (is_resource($data)) {
+            while (!feof($data)) {
+                fwrite($stream, fread($data, 8192));
+            }
+        } else {
+            fwrite($stream, $data);
+        }
+
+        stream_filter_remove($filter_body);
+        stream_filter_remove($filter_data);
+
+        $size = ftell($stream);
+        rewind($stream);
+
+        $chunking = $this->queryExtension('CHUNKING');
+        $chunk_force = false;
+        $chunk_size = $this->getParam('chunk_size');
+
+        /* Do body checks now, since if they fail it doesn't make sense to
+         * do anything else. */
+        $body = null;
+        if ($eai || ($filter_body_params->body === '8bit')) {
+            /* RFC 6152[3]. RFC 6531[1.2] also requires at least 8BITMIME to
+             * be available. */
+            $body = $this->data_8bit
+                ? ' BODY=8BITMIME'
+                : false;
+        }
+
+        /* Fallback to binarymime if 8bitmime is not available. */
+        if (($body === false) || ($filter_body_params->body === 'binary')) {
+            // RFC 3030[3]
+            if (!$this->data_binary) {
+                switch ($filter_body_params->body) {
+                case 'binary':
+                    throw new InvalidArgumentException(
+                        'Server does not support binary message data.'
+                    );
+
+                default:
+                    throw new InvalidArgumentException(
+                        'Server does not support 8-bit message data.'
+                    );
+                }
+            }
+
+            $body = ' BODY=BINARYMIME';
+
+            /* BINARYMIME requires CHUNKING. */
+            $chunking = $chunk_force = true;
+            if (!$chunk_size) {
+                $chunk_size = self::CHUNK_DEFAULT;
+            }
+        }
+
+        if (is_null($body) && $this->_debug->active && $this->data_8bit) {
+            /* Only output extended 7bit command if debug is active (it is
+             * default and does not need to be explicitly declared). */
+            $body = ' BODY=7BIT';
+        }
+
+        $mailcmd = 'MAIL FROM:<' . ($eai ? $from->bare_address : $from->bare_address_idn) . '>';
+
+        // RFC 1870[6]
+        if ($this->queryExtension('SIZE')) {
+            $mailcmd .= ' SIZE=' . intval($size);
+        }
+
+        // RFC 6531[3.4]
+        if ($eai) {
+            $mailcmd .= ' SMTPUTF8';
+        }
+
+        if (!is_null($body)) {
+            $mailcmd .= $body;
+        }
+
+        $recipients = $eai
+            ? $to->bare_addresses
+            : $to->bare_addresses_idn;
+
+        $recip_cmds = array();
         foreach ($recipients as $val) {
-            $cmds[] = 'RCPT TO:<' . $val . '>';
+            $recip_cmds[$val] = 'RCPT TO:<' . $val . '>';
         }
 
         if ($this->queryExtension('PIPELINING')) {
-            $this->_connection->write($cmds);
+            $this->_connection->write(
+                array_merge(array($mailcmd), array_values($recip_cmds))
+            );
 
-            $error = null;
-            foreach ($cmds as $val) {
+            try {
+                $this->_getResponse(250);
+                $error =  null;
+            } catch (Horde_Smtp_Exception $e) {
+                $error = $e;
+            }
+
+            foreach ($recip_cmds as $key => $val) {
                 try {
-                    $this->_getResponse(array(250, 251));
-                } catch (Horde_Smtp_Exception $e) {
+                    $this->_getResponse(array(250, 251), array(
+                        'exception' => 'Horde_Smtp_Exception_Recipients'
+                    ));
+                } catch (Horde_Smtp_Exception_Recipients $e) {
                     if (is_null($error)) {
                         $error = $e;
                     }
+
+                    if ($error instanceof Horde_Smtp_Exception_Recipients) {
+                        $error->recipients[] = $key;
+                    }
                 }
             }
 
@@ -594,48 +666,74 @@ class Horde_Smtp implements Serializable
                 throw $error;
             }
         } else {
-            foreach ($cmds as $val) {
+            $this->_connection->write($mailcmd);
+            $this->_getResponse(250, array(
+                'error' => 'reset'
+            ));
+
+            foreach ($recip_cmds as $key => $val) {
                 $this->_connection->write($val);
-                $this->_getResponse(array(250, 251), 'reset');
-            }
-        }
-
-        $this->_connection->write('DATA');
-
-        try {
-            $this->_getResponse(354, 'reset');
-        } catch (Horde_Smtp_Exception $e) {
-            /* This is the place where a STARTTLS 530 error would occur. If
-             * so, explicitly use STARTTLS and try again. */
-            switch ($e->getSmtpCode()) {
-            case 530:
-                if (!$this->isSecureConnection()) {
-                    $this->logout();
-                    $this->setParam('secure', 'tls');
-                    $this->send($from, $to, $data, $opts);
-                    return;
+                try {
+                    $this->_getResponse(array(250, 251), array(
+                        'error' => 'reset',
+                        'exception' => 'Horde_Smtp_Exception_Recipients'
+                    ));
+                } catch (Horde_Smtp_Exception_Recipients $e) {
+                    $e->recipients[] = $key;
+                    throw $e;
                 }
-                break;
+            }
+        }
+
+        /* CHUNKING support. RFC 3030[2] */
+        if ($chunking &&
+            $chunk_size &&
+            ($chunk_force || ($size > $chunk_size))) {
+            while ($size) {
+                $c = min($chunk_size, $size);
+                $size -= $c;
+
+                $this->_connection->write(
+                    'BDAT ' . $c . ($size ? '' : ' LAST')
+                );
+                $this->_connection->write($stream, $c);
+                if ($size) {
+                    $this->_getResponse(250, array(
+                        'error' => 'reset'
+                    ));
+                }
+            }
+        } else {
+            $this->_connection->write('DATA');
+
+            try {
+                $this->_getResponse(354, array(
+                    'error' => 'reset'
+                ));
+            } catch (Horde_Smtp_Exception $e) {
+                fclose($stream);
+
+                /* This is the place where a STARTTLS 530 error would occur.
+                 * If so, explicitly use STARTTLS and try again. */
+                switch ($e->getSmtpCode()) {
+                case 530:
+                    if (!$this->isSecureConnection()) {
+                        $this->logout();
+                        $this->setParam('secure', 'tls');
+                        $this->send($from, $to, $data, $opts);
+                        return;
+                    }
+                    break;
+                }
+
+                throw $e;
             }
 
-            throw $e;
+            $this->_connection->write($stream);
+            $this->_connection->write('.');
         }
 
-        if (!is_resource($data)) {
-            $stream = fopen('php://temp', 'r+');
-            fwrite($stream, $data);
-            $data = $stream;
-        }
-        rewind($data);
-
-        // Add SMTP escape filter.
-        stream_filter_register('horde_smtp_data', 'Horde_Smtp_Filter_Data');
-        $res = stream_filter_append($data, 'horde_smtp_data', STREAM_FILTER_READ);
-
-        $this->_connection->write($data);
-        stream_filter_remove($res);
-
-        $this->_connection->write('.');
+        fclose($stream);
 
         return $this->_processData($recipients);
     }
@@ -674,6 +772,32 @@ class Horde_Smtp implements Serializable
         }
     }
 
+    /**
+     * Send request to process the remote queue.
+     *
+     * @param string $host  The specific host to request queue processing for.
+     *
+     * @throws Horde_Smtp_Exception
+     */
+    public function processQueue($host = null)
+    {
+        $this->login();
+
+        if (!$this->queryExtension('ETRN')) {
+            return;
+        }
+
+        if (is_null($host)) {
+            $host = $this->_getHostname();
+            if ($host === false) {
+                return;
+            }
+        }
+
+        $this->_connection->write('ETRN ' . $host);
+        $this->_getResponse(array(250, 251, 252, 253));
+    }
+
     /* Internal methods. */
 
     /**
@@ -683,7 +807,7 @@ class Horde_Smtp implements Serializable
      */
     protected function _hello()
     {
-        $ehlo = $host = gethostname();
+        $ehlo = $host = $this->_getHostname();
         if ($host === false) {
             $ehlo = $_SERVER['SERVER_ADDR'];
             $host = 'localhost';
@@ -748,7 +872,9 @@ class Horde_Smtp implements Serializable
         }
 
         $this->_connection->write('STARTTLS');
-        $this->_getResponse(220, 'logout');
+        $this->_getResponse(220, array(
+            'error' => 'logout'
+        ));
 
         if (!$this->_connection->startTls()) {
             $this->logout();
@@ -789,7 +915,7 @@ class Horde_Smtp implements Serializable
 
             $this->_debug->active = false;
             $this->_connection->write(
-                base64_encode($user . ' ' . hash_hmac(strtolower(substr($method, 5)), base64_decode(reset($resp)), $pass, false))
+                base64_encode($user . ' ' . hash_hmac(Horde_String::lower(substr($method, 5)), base64_decode(reset($resp)), $pass, false))
             );
             $this->_debug->active = true;
 
@@ -879,14 +1005,23 @@ class Horde_Smtp implements Serializable
     /**
      * Gets a line from the incoming stream and parses it.
      *
-     * @param mixed $code    Expected reply code(s) (integer or array).
-     * @param string $error  On error, 'logout' or 'reset'?
+     * @param mixed $code  Expected reply code(s) (integer or array).
+     * @param array $opts  Additional options:
+     * <pre>
+     *   - error: (string) On error, 'logout' or 'reset'?
+     *   - exception: (string) Throw an exception of this class on error.
+     * </pre>
      *
      * @return array  An array with the response text.
      * @throws Horde_Smtp_Exception
      */
-    protected function _getResponse($code, $error = null)
+    protected function _getResponse($code, array $opts = array())
     {
+        $opts = array_merge(array(
+            'error' => null,
+            'exception' => 'Horde_Smtp_Exception'
+        ), $opts);
+
         $text = array();
 
         while ($read = $this->_connection->read()) {
@@ -911,16 +1046,22 @@ class Horde_Smtp implements Serializable
         if (!is_null($this->_extensions) &&
             $this->queryExtension('ENHANCEDSTATUSCODES')) {
             list($enhanced, $details) = explode(' ', $details, 2);
+            if (!strpos($enhanced, '.')) {
+                $details = $enhanced . ' ' . $details;
+                $enhanced = null;
+            }
         } else {
             $enhanced = null;
         }
 
-        $e = new Horde_Smtp_Exception($details);
+        $e = new $opts['exception']($details);
         $e->details = $details;
         $e->setSmtpCode($replycode);
-        $e->setEnhancedSmtpCode($enhanced);
+        if ($enhanced) {
+            $e->setEnhancedSmtpCode($enhanced);
+        }
 
-        switch ($error) {
+        switch ($opts['error']) {
         case 'logout':
             $this->logout();
             break;
@@ -948,8 +1089,22 @@ class Horde_Smtp implements Serializable
      */
     protected function _processData($recipients)
     {
-        $this->_getResponse(250, 'reset');
+        $this->_getResponse(250, array(
+            'error' => 'reset'
+        ));
         return array_fill_keys($recipients, true);
     }
 
+    /**
+     * Return the local hostname.
+     *
+     * @return string  Local hostname (null if it cannot be determined).
+     */
+    protected function _getHostname()
+    {
+        return ($localhost = $this->getParam('localhost'))
+            ? $localhost
+            : gethostname();
+    }
+
 }
