@@ -1,12 +1,12 @@
 <?php
 /**
- * Copyright 2013-2014 Horde LLC (http://www.horde.org/)
+ * Copyright 2013-2015 Horde LLC (http://www.horde.org/)
  *
  * See the enclosed file COPYING for license information (LGPL). If you
  * did not receive this file, see http://www.horde.org/licenses/lgpl21.
  *
  * @category  Horde
- * @copyright 2013-2014 Horde LLC
+ * @copyright 2013-2015 Horde LLC
  * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
  * @package   Smtp
  */
@@ -17,7 +17,7 @@
  *
  * @author    Michael Slusarz <slusarz@horde.org>
  * @category  Horde
- * @copyright 2013-2014 Horde LLC
+ * @copyright 2013-2015 Horde LLC
  * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
  * @package   Smtp
  */
@@ -51,9 +51,9 @@ class Horde_Smtp_Filter_Data extends php_user_filter
             }
 
             // EOLs need to be CRLF; double leading periods.
-            $bucket->data = preg_replace(
-                array("/(?:\r\n|\n|\r(?!\n))/", "/\n\./"),
-                array("\r\n", "\n.."),
+            $bucket->data = str_replace(
+                array("\r\n", "\r", "\n", "\n."),
+                array("\n", "\n", "\r\n", "\n.."),
                 $bucket->data
             );
 
