@@ -1,6 +1,6 @@
 <?php
 /**
- * Copyright 2013-2014 Horde LLC (http://www.horde.org/)
+ * Copyright 2013-2015 Horde LLC (http://www.horde.org/)
  *
  * See the enclosed file COPYING for license information (LGPL). If you
  * did not receive this file, see http://www.horde.org/licenses/lgpl21.
@@ -70,6 +70,15 @@ class Horde_Smtp_RemoteServerTest extends Horde_Test_Case
         );
     }
 
+    public function testBinaryMime()
+    {
+        $this->_createSmtp();
+        $this->assertEquals(
+            $this->smtp->queryExtension('BINARYMIME'),
+            $this->smtp->data_binary
+        );
+    }
+
     public function testSize()
     {
         $this->_createSmtp();
@@ -79,6 +88,18 @@ class Horde_Smtp_RemoteServerTest extends Horde_Test_Case
         );
     }
 
+    public function testNoop()
+    {
+        $this->_createSmtp();
+        $this->smtp->noop();
+    }
+
+    public function testProcessQueue()
+    {
+        $this->_createSmtp();
+        $this->smtp->processQueue();
+    }
+
     protected function _createSmtp()
     {
         $this->smtp = new Horde_Smtp($this->config);
